# Copyright 2014 OneConvergence, Inc. All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy
import httplib
import requests
import urlparse

from oslo.config import cfg

from gbpservice.neutron.services.grouppolicy.common import exceptions

from neutron.openstack.common import jsonutils
from neutron.openstack.common import log as logging


SERVICE_CONTROLLER_OPTIONS = [
    cfg.StrOpt('service_controller_ip',
               help=_('One Convergence NVSD Service Controller IP Address')),
    cfg.StrOpt('service_controller_port',
               help=_('One Convergence NVSD Service Controller Port Number')),
    cfg.StrOpt('request_retries',
               help=_('One Convergence NVSD Service Controller API '
                      'request retries')),
    cfg.StrOpt('request_timeout',
               help=_('One Convergence NVSD Service Controller API '
                      'request timeout')),
    cfg.StrOpt('api_version',
               default='1.0',
               help=_('One Convergence NVSD Service Controller API Version')),
]

cfg.CONF.register_opts(SERVICE_CONTROLLER_OPTIONS, "NVSD_SERVICE_CONTROLLER")

LOG = logging.getLogger(__name__)

NVSD_ENDPOINT = "/nvsd_connectivity_port"
NVSD_ENDPOINT_GROUP = "/nvsd_connectivity_portgroup"
NVSD_CONTRACT = "/nvsd_connectivity_contract"
NVSD_POLICY = "/nvsd_connectivity_policy"
NVSD_POLICY_ACTION = "/nvsd_connectivity_action"
NVSD_POLICY_CLASSIFIER = "/nvsd_connectivity_classifier"
NVSD_POLICY_RULE = "/nvsd_connectivity_rule"
NVSD_SERVICE = "/service"

ADMIN_URL = "&is_admin=true"
API_TENANT_USER = "?tenant_id=%s&user_id=%s"


class GroupPolicyException(exceptions.GroupPolicyException):
    """Base for policy driver exceptions returned to user."""
    message = _("Unexpected response code %(status)s from NVSD "
                "Service Controller for %(method)s to %(url)s")


class NVSDServiceController(object):
    """Encapsulates the One Convergence NVSD Service Controller details.

    Uses python-requests library to perform API request to One Convergence
    NVSD Service Controller.
    """

    def __init__(self):

        self._host = cfg.CONF.NVSD_SERVICE_CONTROLLER.service_controller_ip
        self._port = cfg.CONF.NVSD_SERVICE_CONTROLLER.service_controller_port
        self._retries = cfg.CONF.NVSD_SERVICE_CONTROLLER.request_retries
        self._request_timeout = float(cfg.CONF.NVSD_SERVICE_CONTROLLER.
                                      request_timeout)
        self.service_api_url = 'http://' + self._host + ':' + str(self._port)
        self.pool = requests.Session()

    def do_request(self, method, url=None, headers=None, data=None,
                   timeout=10):
        response = self.pool.request(method, url=url,
                                     headers=headers, data=data,
                                     timeout=timeout)
        return response

    def request(self, method, uri, context, body="",
                content_type="application/json", filters={}):
        """Issue a request to NVSD Service Controller."""

        headers = {"Content-Type": content_type}
        api_version = "/v" + cfg.CONF.NVSD_SERVICE_CONTROLLER.api_version
        uri = api_version + uri
        if context.is_admin:
            uri = uri + ADMIN_URL
        if filters.get('tenant_id'):
            uri = uri + "&filter_tenant_id=%s" % filters.get('tenant_id')[0]

        url = urlparse.urljoin(self.service_api_url, uri)

        response = None

        try:
            response = self.do_request(method, url=url, headers=headers,
                                       data=body,
                                       timeout=self._request_timeout)

            LOG.debug("Request: %(method)s %(uri)s executed",
                      {'method': method, 'uri': self.service_api_url + uri})
        except httplib.IncompleteRead as err:
            response = err.partial
        except Exception as err:
            LOG.error(_("Request failed in NVSD Service Controller. "
                        "Error : %s"), err)

        if response is None:
            # Request was timed out.
            LOG.error(_("Response is Null, Request for method : %(method)s to "
                        "%(uri)s Timed out"), {'method': method, 'uri': uri})
            raise GroupPolicyException(status="TimedOut", method=method,
                                       url=self.service_api_url + uri)

        status = response.status_code
        #Not Found (404) is OK for DELETE. Ignore it here
        if method == 'DELETE' and status == 404:
            return
        elif status not in (requests.codes.ok, requests.codes.created,
                            requests.codes.no_content):
            LOG.error(_("Unexpected response code %(status)s from NVSD "
                        "Service Controller for %(method)s to %(url)s"),
                      {'status': status, 'method': method, 'url': url})
            raise GroupPolicyException(status=status, method=method,
                                       url=self.service_api_url + uri)
        else:
            LOG.debug("Success: %(method)s %(url)s status=%(status)s",
                      {'method': method, 'url': self.service_api_url + uri,
                       'status': status})
        response.body = response.content
        return response


class NVSDServiceApi(object):
    """Invokes One Convergence NVSD Service Controller API.

    Invokes the appropriate One Convergence NVSD Service Controller API for
    each of the Openstack Group Based Policy API operation. Maps the Openstack
    Group Policy parameters to One Convergence NVSD API parameters.
    """

    def __init__(self):
        self.nvsd_service_controller = NVSDServiceController()

    def create_policy_classifier(self, context, policy_classifier):
        body = copy.deepcopy(policy_classifier)
        body.update({"port": policy_classifier.get("port_range")})
        tenant_id = context.tenant_id
        uri = (NVSD_POLICY_CLASSIFIER + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request("POST", uri, context,
                                                        jsonutils.dumps(body))
        return response.json()

    def get_policy_classifiers(self, context, tenant_id, filters={}):
        uri = (NVSD_POLICY_CLASSIFIER + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def get_policy_classifier(self, context, classifier_id):
        uri = (NVSD_POLICY_CLASSIFIER + "/%s?tenant_id=%s&user_id=%s" %
               (classifier_id, context.tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def update_policy_classifier(self, context, classifier):
        tenant_id = context.tenant_id
        classifier_id = classifier.get('id')
        body = copy.deepcopy(classifier)
        body.update({"port": classifier.get("port_range")})
        uri = (NVSD_POLICY_CLASSIFIER + "/%s?tenant_id=%s&user_id=%s" %
               (classifier_id, tenant_id, context.user))
        response = self.nvsd_service_controller.request("PUT", uri, context,
                                                        jsonutils.dumps(body))
        return response.json()

    def delete_policy_classifier(self, context, classifier_id):
        tenant_id = context.tenant_id
        uri = (NVSD_POLICY_CLASSIFIER + "/%s?tenant_id=%s&user_id=%s" %
               (classifier_id, tenant_id, context.user))
        self.nvsd_service_controller.request("DELETE", uri, context)

    def create_policy_rule(self, context, rule):
        '''
        body = copy.deepcopy(rule)
        body.update({'classifier': rule.get('policy_classifier_id'),
                     'actions': rule.get('policy_actions', []),
                     'policies_attached': []})
        '''
        tenant_id = context.tenant_id
        uri = (NVSD_POLICY_RULE + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request("POST", uri, context,
                                                        jsonutils.dumps(rule))
        return response.json()

    def update_policy_rule(self, context, rule):
        tenant_id = context.tenant_id
        rule_id = rule.get('id')
        body = copy.deepcopy(rule)
        body.update({'classifier': rule.get('policy_classifier_id'),
                     'actions': rule.get('policy_actions', [])})
        uri = (NVSD_POLICY_RULE + "/%s?tenant_id=%s&user_id=%s" %
               (rule_id, tenant_id, context.user))
        response = self.nvsd_service_controller.request("PUT", uri, context,
                                                        jsonutils.dumps(body))
        return response.json()

    def get_policy_rule(self, context, rule_id):
        uri = (NVSD_POLICY_RULE + "/%s?tenant_id=%s&user_id=%s" %
               (rule_id, context.tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def get_policy_rules(self, context, tenant_id, filters={}):
        uri = (NVSD_POLICY_RULE + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def delete_policy_rule(self, context, rule_id):
        uri = (NVSD_POLICY_RULE + "/%s?tenant_id=%s&user_id=%s" %
               (rule_id, context.tenant_id, context.user))
        self.nvsd_service_controller.request("DELETE", uri, context)

    def create_policy_action(self, context, action):
        body = copy.deepcopy(action)
        action_type = action.get("action_type")
        if action_type.lower() == "redirect":
            body["action_type"] = "l2redirect"
        tenant_id = context.tenant_id
        uri = (NVSD_POLICY_ACTION + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request("POST", uri, context,
                                                        jsonutils.dumps(body))
        return response.json()

    def update_policy_action(self, context, policy_action):
        tenant_id = context.tenant_id
        action_id = policy_action.get('id')
        body = copy.deepcopy(policy_action)
        action_type = policy_action.get("action_type")
        if action_type.lower() == "redirect":
            body["action_type"] = "l2redirect"
        uri = (NVSD_POLICY_ACTION + "/%s?tenant_id=%s&user_id=%s" %
               (action_id, tenant_id, context.user))
        response = self.nvsd_service_controller.request("PUT", uri, context,
                                                        jsonutils.dumps(body))
        return response.json()

    def get_policy_action(self, context, action_id):
        uri = (NVSD_POLICY_ACTION + "/%s?tenant_id=%s&user_id=%s" %
               (action_id, context.tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri)
        return response.json()

    def get_policy_actions(self, context, tenant_id, filters={}):
        uri = (NVSD_POLICY_ACTION + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def delete_policy_action(self, context, action_id):
        uri = (NVSD_POLICY_ACTION + "/%s?tenant_id=%s&user_id=%s" %
               (action_id, context.tenant_id, context.user))
        self.nvsd_service_controller.request("DELETE", uri, context)

    def create_endpointgroup(self, context, endpointgroup):
        uri = (NVSD_ENDPOINT_GROUP + "?tenant_id=%s&user_id=%s" %
               (context.tenant_id, context.user))
        response = self.nvsd_service_controller.request(
                            "POST", uri, context,
                            jsonutils.dumps(endpointgroup))

        return response.json()

    def get_endpointgroups(self, context, tenant_id, filters={}):
        uri = (NVSD_ENDPOINT_GROUP + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def get_endpointgroup(self, context, endpointgroup_id):
        uri = (NVSD_ENDPOINT_GROUP + "/%s?tenant_id=%s&user_id=%s" %
               (endpointgroup_id, context.tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def update_endpointgroup(self, context, endpointgroup):
        tenant_id = context.tenant_id
        endpointgroup_id = endpointgroup.get('id')
        uri = (NVSD_ENDPOINT_GROUP + "/%s?tenant_id=%s&user_id=%s" %
               (endpointgroup_id, tenant_id, context.user))
        response = self.nvsd_service_controller.request(
                                "PUT", uri, context,
                                jsonutils.dumps(endpointgroup))
        return response.json()

    def delete_endpointgroup(self, context, endpointgroup_id):
        uri = (NVSD_ENDPOINT_GROUP + "/%s?tenant_id=%s&user_id=%s" %
               (endpointgroup_id, context.tenant_id, context.user))
        self.nvsd_service_controller.request("DELETE", uri, context)

    def create_endpoint(self, context, endpoint):
        body = copy.deepcopy(endpoint)
        body.update({'connectivity_portgroup_id':
                     endpoint.get('policy_target_group_id')})

        tenant_id = context.tenant_id
        uri = (NVSD_ENDPOINT + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request("POST", uri, context,
                                                        jsonutils.dumps(body))
        return response.json()

    def update_endpoint(self, context, endpoint):
        tenant_id = context.tenant_id
        endpoint_id = endpoint.get('id')
        body = copy.deepcopy(endpoint)
        body.update({'connectivity_portgroup_id':
                     endpoint.get('policy_target_group_id')})

        uri = (NVSD_ENDPOINT + "/%s?tenant_id=%s&user_id=%s" %
               (endpoint_id, tenant_id, context.user))
        response = self.nvsd_service_controller.request("PUT", uri, context,
                                                        jsonutils.dumps(body))
        return response.json()

    def get_endpoint(self, context, endpoint_id):
        uri = (NVSD_ENDPOINT + "/%s?tenant_id=%s&user_id=%s" %
               (endpoint_id, context.tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def get_endpoints(self, context, tenant_id, filters={}):
        uri = (NVSD_ENDPOINT + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def delete_endpoint(self, context, endpoint_id):
        uri = (NVSD_ENDPOINT + "/%s?tenant_id=%s&user_id=%s" %
               (endpoint_id, context.tenant_id, context.user))
        self.nvsd_service_controller.request("DELETE", uri, context)
        return

    def create_contract(self, context, contract):
        tenant_id = context.tenant_id
        uri = (NVSD_CONTRACT + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request(
                            "POST", uri, context, jsonutils.dumps(contract))
        return response.json()

    def update_contract(self, context, contract):
        tenant_id = context.tenant_id
        contract_id = contract.get('id')
        uri = (NVSD_CONTRACT + "/%s?tenant_id=%s&user_id=%s" %
               (contract_id, tenant_id, context.user))
        response = self.nvsd_service_controller.request(
                                "PUT", uri, context, jsonutils.dumps(contract))
        return response.json()

    def get_contract(self, context, contract_id):
        uri = (NVSD_CONTRACT + "/%s?tenant_id=%s&user_id=%s" %
               (contract_id, context.tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def get_contracts(self, context, tenant_id, filters={}):
        uri = (NVSD_CONTRACT + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def delete_contract(self, context, contract_id):
        uri = (NVSD_CONTRACT + "/%s?tenant_id=%s&user_id=%s" %
               (contract_id, context.tenant_id, context.user))
        self.nvsd_service_controller.request("DELETE", uri, context)

    def create_policy(self, context, policy):
        tenant_id = context.tenant_id
        uri = (NVSD_POLICY + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request(
                                "POST", uri, context, jsonutils.dumps(policy))
        return response.json()

    def update_policy(self, context, policy):
        tenant_id = context.tenant_id
        policy_id = policy.get('id')
        uri = (NVSD_POLICY + "/%s?tenant_id=%s&user_id=%s" %
               (policy_id, tenant_id, context.user))
        response = self.nvsd_service_controller.request(
                                "PUT", uri, context, jsonutils.dumps(policy))
        return response.json()

    def get_policy(self, context, policy_id):
        uri = (NVSD_POLICY + "/%s?tenant_id=%s&user_id=%s" %
               (policy_id, context.tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def get_policys(self, context, tenant_id, filters={}):
        uri = (NVSD_POLICY + "?tenant_id=%s&user_id=%s" %
               (tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()

    def delete_policy(self, context, policy_id):
        uri = (NVSD_POLICY + "/%s?tenant_id=%s&user_id=%s" %
               (policy_id, context.tenant_id, context.user))
        self.nvsd_service_controller.request("DELETE", uri, context)

    def get_nvsd_service(self, context, service_id):
        uri = (NVSD_SERVICE + "/%s?tenant_id=%s&user_id=%s" %
               (service_id, context.tenant_id, context.user))
        response = self.nvsd_service_controller.request("GET", uri, context)
        return response.json()



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import webob.exc

from neutron.api import extensions
from neutron import context
from neutron.openstack.common import importutils
from neutron.openstack.common import uuidutils
from neutron.plugins.common import constants
from neutron.tests.unit import test_db_plugin
from neutron.tests.unit import test_extensions

from gbpservice.neutron.db import servicechain_db as svcchain_db
import gbpservice.neutron.extensions
from gbpservice.neutron.extensions import servicechain as service_chain

JSON_FORMAT = 'json'


class ServiceChainDBTestBase(object):
    resource_prefix_map = dict(
        (k, constants.COMMON_PREFIXES[constants.SERVICECHAIN])
        for k in service_chain.RESOURCE_ATTRIBUTE_MAP.keys()
    )

    fmt = JSON_FORMAT

    def _get_resource_plural(self, resource):
        if resource.endswith('y'):
            resource_plural = resource.replace('y', 'ies')
        else:
            resource_plural = resource + 's'

        return resource_plural

    def _test_list_resources(self, resource, items,
                             neutron_context=None,
                             query_params=None):
        resource_plural = self._get_resource_plural(resource)

        res = self._list(resource_plural,
                         neutron_context=neutron_context,
                         query_params=query_params)
        resource = resource.replace('-', '_')
        self.assertEqual(sorted([i['id'] for i in res[resource_plural]]),
                         sorted([i[resource]['id'] for i in items]))

    def _get_test_servicechain_node_attrs(self, name='scn1',
                                          description='test scn',
                                          service_type=constants.FIREWALL,
                                          config="{}"):
        attrs = {'name': name, 'description': description,
                 'service_type': service_type,
                 'config': config,
                 'tenant_id': self._tenant_id}

        return attrs

    def _get_test_servicechain_spec_attrs(self, name='scs1',
                                          description='test scs',
                                          nodes=None):
        node_ids = []
        if nodes:
            node_ids = [node_id for node_id in nodes]
        attrs = {'name': name, 'description': description,
                 'tenant_id': self._tenant_id,
                 'nodes': node_ids}

        return attrs

    def _get_test_servicechain_instance_attrs(self, name='sci1',
                                              description='test sci',
                                              config_param_values="{}",
                                              servicechain_specs=[],
                                              provider_ptg_id=None,
                                              consumer_ptg_id=None,
                                              classifier_id=None):
        attrs = {'name': name, 'description': description,
                 'tenant_id': self._tenant_id,
                 'config_param_values': config_param_values,
                 'servicechain_specs': servicechain_specs,
                 'provider_ptg_id': provider_ptg_id,
                 'consumer_ptg_id': consumer_ptg_id,
                 'classifier_id': classifier_id}

        return attrs

    def create_servicechain_node(self, service_type=constants.FIREWALL,
                                 config="{}", expected_res_status=None,
                                 **kwargs):
        defaults = {'name': 'scn1', 'description': 'test scn'}
        defaults.update(kwargs)

        data = {'servicechain_node': {'service_type': service_type,
                                      'tenant_id': self._tenant_id,
                                      'config': config}}
        data['servicechain_node'].update(defaults)

        scn_req = self.new_create_request('servicechain_nodes', data, self.fmt)
        scn_res = scn_req.get_response(self.ext_api)

        if expected_res_status:
            self.assertEqual(scn_res.status_int, expected_res_status)
        elif scn_res.status_int >= webob.exc.HTTPClientError.code:
            raise webob.exc.HTTPClientError(code=scn_res.status_int)

        scn = self.deserialize(self.fmt, scn_res)

        return scn

    def create_servicechain_spec(self, nodes=None, expected_res_status=None,
                                 **kwargs):
        defaults = {'name': 'scs1', 'description': 'test scs'}
        defaults.update(kwargs)

        data = {'servicechain_spec': {'tenant_id': self._tenant_id,
                                      'nodes': nodes}}
        data['servicechain_spec'].update(defaults)

        scs_req = self.new_create_request('servicechain_specs', data, self.fmt)
        scs_res = scs_req.get_response(self.ext_api)

        if expected_res_status:
            self.assertEqual(scs_res.status_int, expected_res_status)
        elif scs_res.status_int >= webob.exc.HTTPClientError.code:
            raise webob.exc.HTTPClientError(code=scs_res.status_int)

        scs = self.deserialize(self.fmt, scs_res)

        return scs

    def test_create_servicechain_specs_same_node(self):
        template1 = '{"key1":"value1"}'
        scn = self.create_servicechain_node(config=template1)
        scn_id = scn['servicechain_node']['id']
        spec1 = {"servicechain_spec": {'name': 'scs1',
                                       'tenant_id': self._tenant_id,
                                       'nodes': [scn_id]}}
        spec_req = self.new_create_request('servicechain_specs',
                                           spec1,
                                           self.fmt)
        spec_res = spec_req.get_response(self.ext_api)
        self.assertEqual(spec_res.status_int, webob.exc.HTTPCreated.code)
        res = self.deserialize(self.fmt, spec_res)
        self.assertIn('servicechain_spec', res)
        self.assertEqual([scn_id], res['servicechain_spec']['nodes'])
        spec2 = {"servicechain_spec": {'name': 'scs2',
                                       'tenant_id': self._tenant_id,
                                       'nodes': [scn_id]}}
        spec_req = self.new_create_request('servicechain_specs',
                                           spec2,
                                           self.fmt)
        spec_res = spec_req.get_response(self.ext_api)
        self.assertEqual(spec_res.status_int, webob.exc.HTTPCreated.code)
        res = self.deserialize(self.fmt, spec_res)
        self.assertIn('servicechain_spec', res)
        self.assertEqual([scn_id], res['servicechain_spec']['nodes'])

    def create_servicechain_instance(self, servicechain_specs=[],
                                     config_param_values="{}",
                                     provider_ptg_id=None,
                                     consumer_ptg_id=None,
                                     classifier_id=None,
                                     expected_res_status=None, **kwargs):
        defaults = {'name': 'sci1', 'description': 'test sci'}
        defaults.update(kwargs)
        data = {'servicechain_instance':
                {'config_param_values': config_param_values,
                 'servicechain_specs': servicechain_specs,
                 'tenant_id': self._tenant_id,
                 'provider_ptg_id': provider_ptg_id,
                 'consumer_ptg_id': consumer_ptg_id,
                 'classifier_id': classifier_id}}
        data['servicechain_instance'].update(defaults)

        sci_req = self.new_create_request('servicechain_instances',
                                          data, self.fmt)
        sci_res = sci_req.get_response(self.ext_api)

        if expected_res_status:
            self.assertEqual(sci_res.status_int, expected_res_status)
        elif sci_res.status_int >= webob.exc.HTTPClientError.code:
            raise webob.exc.HTTPClientError(code=sci_res.status_int)

        sci = self.deserialize(self.fmt, sci_res)

        return sci


class ServiceChainDBTestPlugin(svcchain_db.ServiceChainDbPlugin):

        supported_extension_aliases = ['servicechain']


DB_GP_PLUGIN_KLASS = (ServiceChainDBTestPlugin.__module__ + '.' +
                      ServiceChainDBTestPlugin.__name__)

GP_PLUGIN_KLASS = (
    "gbpservice.neutron.services.grouppolicy.plugin.GroupPolicyPlugin")


class ServiceChainDbTestCase(ServiceChainDBTestBase,
                             test_db_plugin.NeutronDbPluginV2TestCase):

    def setUp(self, core_plugin=None, sc_plugin=None, service_plugins=None,
              ext_mgr=None):
        extensions.append_api_extensions_path(
            gbpservice.neutron.extensions.__path__)
        if not sc_plugin:
            sc_plugin = DB_GP_PLUGIN_KLASS
        self.plugin = importutils.import_object(sc_plugin)
        if not service_plugins:
            service_plugins = {'gp_plugin_name': GP_PLUGIN_KLASS,
                               'sc_plugin_name': sc_plugin}

        super(ServiceChainDbTestCase, self).setUp(
            plugin=core_plugin, ext_mgr=ext_mgr,
            service_plugins=service_plugins
        )

        if not ext_mgr:
            ext_mgr = extensions.PluginAwareExtensionManager.get_instance()
            self.ext_api = test_extensions.setup_extensions_middleware(ext_mgr)


class TestServiceChainResources(ServiceChainDbTestCase):

    def _test_show_resource(self, resource, resource_id, attrs):
        resource_plural = self._get_resource_plural(resource)
        req = self.new_show_request(resource_plural, resource_id,
                                    fmt=self.fmt)
        res = self.deserialize(self.fmt,
                               req.get_response(self.ext_api))

        for k, v in attrs.iteritems():
            self.assertEqual(res[resource][k], v)

    def test_create_and_show_servicechain_node(self):
        attrs = self._get_test_servicechain_node_attrs(
            service_type=constants.LOADBALANCER, config="config1")

        scn = self.create_servicechain_node(
            service_type=constants.LOADBALANCER, config="config1")

        for k, v in attrs.iteritems():
            self.assertEqual(scn['servicechain_node'][k], v)

        self._test_show_resource('servicechain_node',
                                 scn['servicechain_node']['id'],
                                 attrs)

    def test_list_servicechain_nodes(self):
        scns = [self.create_servicechain_node(name='scn1', description='scn'),
                self.create_servicechain_node(name='scn2', description='scn'),
                self.create_servicechain_node(name='scn3', description='scn')]
        self._test_list_resources('servicechain_node', scns,
                                  query_params='description=scn')

    def test_update_servicechain_node(self):
        name = 'new_servicechain_node'
        description = 'new desc'
        attrs = self._get_test_servicechain_node_attrs(name=name,
                                                       description=description)

        scn = self.create_servicechain_node()

        data = {'servicechain_node': {'name': name,
                                      'description': description}}
        req = self.new_update_request('servicechain_nodes', data,
                                      scn['servicechain_node']['id'])
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))

        for k, v in attrs.iteritems():
            self.assertEqual(res['servicechain_node'][k], v)

        self._test_show_resource('servicechain_node',
                                 scn['servicechain_node']['id'],
                                 attrs)

    def test_delete_servicechain_node(self):
        ctx = context.get_admin_context()

        scn = self.create_servicechain_node()
        scn_id = scn['servicechain_node']['id']

        scs = self.create_servicechain_spec(nodes=[scn_id])
        scs_id = scs['servicechain_spec']['id']

        # Deleting Service Chain Node in use by a Spec should fail
        self.assertRaises(service_chain.ServiceChainNodeInUse,
                          self.plugin.delete_servicechain_node, ctx, scn_id)

        req = self.new_delete_request('servicechain_specs', scs_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, webob.exc.HTTPNoContent.code)

        # After deleting the Service Chain Spec, node delete should succeed
        req = self.new_delete_request('servicechain_nodes', scn_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, webob.exc.HTTPNoContent.code)
        self.assertRaises(service_chain.ServiceChainNodeNotFound,
                          self.plugin.get_servicechain_node,
                          ctx, scn_id)

    def test_create_and_show_servicechain_spec(self):
        name = "scs1"
        scn = self.create_servicechain_node()
        scn_id = scn['servicechain_node']['id']

        attrs = self._get_test_servicechain_spec_attrs(name, nodes=[scn_id])

        scs = self.create_servicechain_spec(name=name, nodes=[scn_id])

        for k, v in attrs.iteritems():
            self.assertEqual(scs['servicechain_spec'][k], v)

        self._test_show_resource('servicechain_spec',
                                 scs['servicechain_spec']['id'],
                                 attrs)

    def test_create_spec_multiple_nodes(self):
        name = "scs1"
        scn1 = self.create_servicechain_node()
        scn1_id = scn1['servicechain_node']['id']
        scn2 = self.create_servicechain_node()
        scn2_id = scn2['servicechain_node']['id']
        attrs = self._get_test_servicechain_spec_attrs(
                            name, nodes=[scn1_id, scn2_id])
        scs = self.create_servicechain_spec(
                            name=name, nodes=[scn1_id, scn2_id])
        for k, v in attrs.iteritems():
            self.assertEqual(scs['servicechain_spec'][k], v)

    def test_list_servicechain_specs(self):
        scs = [self.create_servicechain_spec(name='scs1', description='scs'),
               self.create_servicechain_spec(name='scs2', description='scs'),
               self.create_servicechain_spec(name='scs3', description='scs')]
        self._test_list_resources('servicechain_spec', scs,
                                  query_params='description=scs')

    def test_node_ordering_list_servicechain_specs(self):
        scn1_id = self.create_servicechain_node()['servicechain_node']['id']
        scn2_id = self.create_servicechain_node()['servicechain_node']['id']
        nodes_list = [scn1_id, scn2_id]
        scs = self.create_servicechain_spec(name='scs1',
                                            nodes=nodes_list)
        self.assertEqual(scs['servicechain_spec']['nodes'], nodes_list)
        res = self._list('servicechain_specs')
        self.assertEqual(len(res['servicechain_specs']), 1)
        self.assertEqual(res['servicechain_specs'][0]['nodes'],
                         nodes_list)

        # Delete the service chain spec and create another with nodes in
        # reverse order and verify that that proper ordering is maintained
        req = self.new_delete_request('servicechain_specs',
                                      scs['servicechain_spec']['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, webob.exc.HTTPNoContent.code)

        nodes_list.reverse()
        scs = self.create_servicechain_spec(name='scs1',
                                            nodes=nodes_list)
        self.assertEqual(scs['servicechain_spec']['nodes'], nodes_list)
        res = self._list('servicechain_specs')
        self.assertEqual(len(res['servicechain_specs']), 1)
        self.assertEqual(res['servicechain_specs'][0]['nodes'],
                         nodes_list)

    def test_update_servicechain_spec(self):
        name = "new_servicechain_spec1"
        description = 'new desc'
        scn_id = self.create_servicechain_node()['servicechain_node']['id']
        attrs = self._get_test_servicechain_spec_attrs(name=name,
                                                       description=description,
                                                       nodes=[scn_id])
        scs = self.create_servicechain_spec()
        data = {'servicechain_spec': {'name': name, 'description': description,
                                      'nodes': [scn_id]}}
        req = self.new_update_request('servicechain_specs', data,
                                      scs['servicechain_spec']['id'])
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))

        for k, v in attrs.iteritems():
            self.assertEqual(res['servicechain_spec'][k], v)

        self._test_show_resource('servicechain_spec',
                                 scs['servicechain_spec']['id'], attrs)

    def test_delete_servicechain_spec(self):
        ctx = context.get_admin_context()

        scs = self.create_servicechain_spec()
        scs_id = scs['servicechain_spec']['id']

        req = self.new_delete_request('servicechain_specs', scs_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, webob.exc.HTTPNoContent.code)
        self.assertRaises(service_chain.ServiceChainSpecNotFound,
                          self.plugin.get_servicechain_spec, ctx, scs_id)

    def test_delete_spec_in_use_by_policy_action_rejected(self):
        ctx = context.get_admin_context()
        scs_id = self.create_servicechain_spec()['servicechain_spec']['id']
        data = {'policy_action': {'action_type': 'redirect',
                                  'tenant_id': self._tenant_id,
                                  'action_value': scs_id}}
        pa_req = self.new_create_request('grouppolicy/policy_actions',
                                         data, self.fmt)
        res = pa_req.get_response(self.ext_api)
        if res.status_int >= webob.exc.HTTPClientError.code:
            raise webob.exc.HTTPClientError(code=res.status_int)

        self.assertRaises(service_chain.ServiceChainSpecInUse,
                          self.plugin.delete_servicechain_spec, ctx, scs_id)

    def test_delete_spec_in_use_by_instance_rejected(self):
        ctx = context.get_admin_context()
        scs_id = self.create_servicechain_spec()['servicechain_spec']['id']

        sci = self.create_servicechain_instance(servicechain_specs=[scs_id])
        sci_id = sci['servicechain_instance']['id']

        # Deleting the Spec used by Instance should not be allowed
        self.assertRaises(service_chain.ServiceChainSpecInUse,
                          self.plugin.delete_servicechain_spec, ctx, scs_id)

        req = self.new_delete_request('servicechain_instances', sci_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, webob.exc.HTTPNoContent.code)
        self.assertRaises(service_chain.ServiceChainInstanceNotFound,
                          self.plugin.get_servicechain_instance,
                          ctx, sci_id)

        # Deleting the spec should succeed after the instance is deleted
        req = self.new_delete_request('servicechain_specs', scs_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, webob.exc.HTTPNoContent.code)
        self.assertRaises(service_chain.ServiceChainSpecNotFound,
                          self.plugin.get_servicechain_spec, ctx, scs_id)

    def test_create_and_show_servicechain_instance(self):
        scs_id = self.create_servicechain_spec()['servicechain_spec']['id']
        policy_target_group_id = uuidutils.generate_uuid()
        classifier_id = uuidutils.generate_uuid()
        config_param_values = "{}"
        attrs = self._get_test_servicechain_instance_attrs(
            servicechain_specs=[scs_id],
            provider_ptg_id=policy_target_group_id,
            consumer_ptg_id=policy_target_group_id,
            classifier_id=classifier_id,
            config_param_values=config_param_values)

        sci = self.create_servicechain_instance(
            servicechain_specs=[scs_id],
            provider_ptg_id=policy_target_group_id,
            consumer_ptg_id=policy_target_group_id,
            classifier_id=classifier_id,
            config_param_values=config_param_values)
        for k, v in attrs.iteritems():
            self.assertEqual(sci['servicechain_instance'][k], v)

        self._test_show_resource('servicechain_instance',
                                 sci['servicechain_instance']['id'],
                                 attrs)
        req = self.new_delete_request('servicechain_instances',
                                      sci['servicechain_instance']['id'])
        req.get_response(self.ext_api)

    def test_list_servicechain_instances(self):
        servicechain_instances = [self.create_servicechain_instance(
            name='sci1', description='sci'),
            self.create_servicechain_instance(name='sci2', description='sci'),
            self.create_servicechain_instance(name='sci3', description='sci')]
        self._test_list_resources('servicechain_instance',
                                  servicechain_instances,
                                  query_params='description=sci')

    def test_spec_ordering_list_servicechain_instances(self):
        scs1_id = self.create_servicechain_spec()['servicechain_spec']['id']
        scs2_id = self.create_servicechain_spec()['servicechain_spec']['id']
        specs_list = [scs1_id, scs2_id]
        sci = self.create_servicechain_instance(name='sci1',
                                                servicechain_specs=specs_list)
        self.assertEqual(sci['servicechain_instance']['servicechain_specs'],
                         specs_list)
        res = self._list('servicechain_instances')
        self.assertEqual(len(res['servicechain_instances']), 1)
        result_instance = res['servicechain_instances'][0]
        self.assertEqual(result_instance['servicechain_specs'], specs_list)

        # Delete the service chain instance and create another with specs in
        # reverse order and verify that that proper ordering is maintained
        req = self.new_delete_request('servicechain_instances',
                                      sci['servicechain_instance']['id'])
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, webob.exc.HTTPNoContent.code)

        specs_list.reverse()
        sci = self.create_servicechain_instance(name='sci1',
                                                servicechain_specs=specs_list)
        self.assertEqual(sci['servicechain_instance']['servicechain_specs'],
                         specs_list)
        res = self._list('servicechain_instances')
        self.assertEqual(len(res['servicechain_instances']), 1)
        result_instance = res['servicechain_instances'][0]
        self.assertEqual(result_instance['servicechain_specs'], specs_list)

    def test_update_servicechain_instance(self):
        name = "new_servicechain_instance"
        description = 'new desc'
        config_param_values = "{}"
        scs_id = self.create_servicechain_spec()['servicechain_spec']['id']
        provider_ptg_id = uuidutils.generate_uuid()
        consumer_ptg_id = uuidutils.generate_uuid()
        classifier_id = uuidutils.generate_uuid()
        attrs = self._get_test_servicechain_instance_attrs(
            name=name, description=description, servicechain_specs=[scs_id],
            provider_ptg_id=provider_ptg_id, consumer_ptg_id=consumer_ptg_id,
            classifier_id=classifier_id,
            config_param_values=config_param_values)

        sci = self.create_servicechain_instance(
            servicechain_specs=[scs_id], provider_ptg_id=provider_ptg_id,
            consumer_ptg_id=consumer_ptg_id, classifier_id=classifier_id,
            config_param_values=config_param_values)
        data = {'servicechain_instance': {'name': name,
                                          'description': description,
                                          'servicechain_specs': [scs_id]}}
        req = self.new_update_request('servicechain_instances', data,
                                      sci['servicechain_instance']['id'])
        res = self.deserialize(self.fmt, req.get_response(self.ext_api))
        for k, v in attrs.iteritems():
            self.assertEqual(res['servicechain_instance'][k], v)

        self._test_show_resource('servicechain_instance',
                                 sci['servicechain_instance']['id'], attrs)
        req = self.new_delete_request('servicechain_instances',
                                      sci['servicechain_instance']['id'])
        req.get_response(self.ext_api)

    def test_delete_servicechain_instance(self):
        ctx = context.get_admin_context()

        sci = self.create_servicechain_instance()
        sci_id = sci['servicechain_instance']['id']

        req = self.new_delete_request('servicechain_instances', sci_id)
        res = req.get_response(self.ext_api)
        self.assertEqual(res.status_int, webob.exc.HTTPNoContent.code)
        self.assertRaises(service_chain.ServiceChainInstanceNotFound,
                          self.plugin.get_servicechain_instance,
                          ctx, sci_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo.config import cfg
import sqlalchemy as sa

from neutron.common import log
from neutron.db import model_base
from neutron.openstack.common import log as logging

from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)
from gbpservice.neutron.services.grouppolicy.common import exceptions as exc


LOG = logging.getLogger(__name__)

opts = [
    cfg.StrOpt('default_l3_policy_name',
               default='default',
               help=_("Name of each tenant's default L3 policy.")),
    cfg.IntOpt('default_ip_version',
               default=4,
               help=_("IP version (4 or 6) for implicitly created default L3 "
                      "policies.")),
    cfg.StrOpt('default_ip_pool',
               default='10.0.0.0/8',
               help=_("IP pool for implicitly created default L3 policies, "
                      "from which subnets are allocated for policy target "
                      "groups.")),
    cfg.IntOpt('default_subnet_prefix_length',
               default=26,
               help=_("Subnet prefix length for implicitly created default L3 "
                      "polices, controlling size of subnets allocated for "
                      "policy target groups.")),

    cfg.StrOpt('default_external_segment_name',
               default='default',
               help=_("Name of default External Segment. This will be used "
                      "whenever a new EP/L3P is created without a referenced "
                      "External Segment. Set to None if a completely "
                      "explicit workflow is preferred.")),
]

cfg.CONF.register_opts(opts, "group_policy_implicit_policy")


class OwnedL2Policy(model_base.BASEV2):
    """An L2 Policy owned by the mapping driver."""

    __tablename__ = 'gpm_owned_l2_policies'
    l2_policy_id = sa.Column(sa.String(36),
                             sa.ForeignKey('gp_l2_policies.id',
                                           ondelete='CASCADE'),
                             nullable=False, primary_key=True)


class OwnedL3Policy(model_base.BASEV2):
    """An L3 Policy owned by the mapping driver."""

    __tablename__ = 'gpm_owned_l3_policies'
    l3_policy_id = sa.Column(sa.String(36),
                             sa.ForeignKey('gp_l3_policies.id',
                                           ondelete='CASCADE'),
                             nullable=False, primary_key=True)


class ImplicitPolicyDriver(api.PolicyDriver):
    """Implicit Policy driver for Group Policy plugin.

    This driver ensures that the l2_policy_id attribute of
    PolicyTargetGroup references an L2Policy instance and that the
    l3_policy_id attribute of L2Policy references an L3Policy instance
    when the default value of None is specified.
    """

    @log.log
    def initialize(self):
        gpip = cfg.CONF.group_policy_implicit_policy
        self._default_l3p_name = gpip.default_l3_policy_name
        self._default_ip_version = gpip.default_ip_version
        self._default_ip_pool = gpip.default_ip_pool
        self._default_subnet_prefix_length = gpip.default_subnet_prefix_length

        self._default_es_name = gpip.default_external_segment_name

    @log.log
    def create_policy_target_group_postcommit(self, context):
        if not context.current['l2_policy_id']:
            self._use_implicit_l2_policy(context)

    @log.log
    def update_policy_target_group_postcommit(self, context):
        old_l2p_id = context.original['l2_policy_id']
        new_l2p_id = context.current['l2_policy_id']
        if old_l2p_id != new_l2p_id:
            self._cleanup_l2_policy(context, old_l2p_id)
            if not new_l2p_id:
                self._use_implicit_l2_policy(context)

    @log.log
    def delete_policy_target_group_postcommit(self, context):
        l2p_id = context.current['l2_policy_id']
        self._cleanup_l2_policy(context, l2p_id)

    @log.log
    def create_l2_policy_postcommit(self, context):
        if not context.current['l3_policy_id']:
            self._use_implicit_l3_policy(context)

    @log.log
    def update_l2_policy_postcommit(self, context):
        old_l3p_id = context.original['l3_policy_id']
        new_l3p_id = context.current['l3_policy_id']
        if old_l3p_id != new_l3p_id:
            self._cleanup_l3_policy(context, old_l3p_id)
            if not new_l3p_id:
                self._use_implicit_l3_policy(context)

    @log.log
    def delete_l2_policy_postcommit(self, context):
        l3p_id = context.current['l3_policy_id']
        self._cleanup_l3_policy(context, l3p_id)

    @log.log
    def create_external_segment_precommit(self, context):
        # REVISIT(ivar): find a better way to retrieve the default ES
        if self._default_es_name == context.current['name']:
            filters = {'name': [self._default_es_name]}
            ess = context._plugin.get_external_segments(
                context._plugin_context, filters)
            if [x for x in ess if x['id'] != context.current['id']]:
                raise exc.DefaultExternalSegmentAlreadyExists(
                    es_name=self._default_es_name)

    @log.log
    def create_external_policy_postcommit(self, context):
        if not context.current['external_segments']:
            self._use_implicit_external_segment(context)

    @log.log
    def update_external_policy_postcommit(self, context):
        old_es_ids = set(context.original['external_segments'])
        new_es_ids = set(context.current['external_segments'])
        added = new_es_ids - old_es_ids
        if not added:
            self._use_implicit_external_segment(context)

    @log.log
    def create_l3_policy_postcommit(self, context):
        if not context.current['external_segments']:
            self._use_implicit_external_segment(context)

    @log.log
    def update_l3_policy_postcommit(self, context):
        old_es_ids = set(context.original['external_segments'].keys())
        new_es_ids = set(context.current['external_segments'].keys())
        added = new_es_ids - old_es_ids
        if not added:
            self._use_implicit_external_segment(context)

    def _use_implicit_l2_policy(self, context):
        attrs = {'l2_policy':
                 {'tenant_id': context.current['tenant_id'],
                  'name': context.current['name'],
                  'description': _("Implicitly created L2 policy"),
                  'l3_policy_id': None,
                  'shared': context.current.get('shared', False),
                  'network_id': None}}
        l2p = context._plugin.create_l2_policy(context._plugin_context, attrs)
        l2p_id = l2p['id']
        self._mark_l2_policy_owned(context._plugin_context.session, l2p_id)
        context.set_l2_policy_id(l2p_id)

    def _cleanup_l2_policy(self, context, l2p_id):
        if self._l2_policy_is_owned(context._plugin_context.session, l2p_id):
            context._plugin.delete_l2_policy(context._plugin_context, l2p_id)

    def _use_implicit_l3_policy(self, context):
        filter = {'tenant_id': [context.current['tenant_id']],
                  'name': [self._default_l3p_name]}
        l3ps = context._plugin.get_l3_policies(context._plugin_context, filter)
        l3p = l3ps and l3ps[0]
        if not l3p:
            # REVISIT(rkukura): Concurrency could result in multiple
            # default L3Ps for the same tenant. A DB table mapping
            # tenant_id to default l3_policy_id may be needed to
            # ensure a single default L3 policy is used per tenant.
            attrs = {'l3_policy':
                     {'tenant_id': context.current['tenant_id'],
                      'name': self._default_l3p_name,
                      'description': _("Implicitly created L3 policy"),
                      'ip_version': self._default_ip_version,
                      'ip_pool': self._default_ip_pool,
                      'shared': context.current.get('shared', False),
                      'subnet_prefix_length':
                      self._default_subnet_prefix_length}}
            l3p = context._plugin.create_l3_policy(context._plugin_context,
                                                   attrs)
            self._mark_l3_policy_owned(context._plugin_context.session,
                                       l3p['id'])
        context.current['l3_policy_id'] = l3p['id']
        context.set_l3_policy_id(l3p['id'])

    def _use_implicit_external_segment(self, context):
        if not self._default_es_name:
            return

        filter = {'name': [self._default_es_name]}
        ess = context._plugin.get_external_segments(context._plugin_context,
                                                    filter)
        # Multiple default ES may exist, this can happen when a per-tenant
        # default ES gets his shared attribute flipped. Always prefer the
        # specific tenant's ES if any.
        for es in ess:
            if es['tenant_id'] == context.current['tenant_id']:
                default = es
                break
        else:
            default = ess and ess[0]
        if default:
            # Set default ES
            context.set_external_segment(default['id'])

    def _cleanup_l3_policy(self, context, l3p_id):
        if self._l3_policy_is_owned(context._plugin_context.session, l3p_id):
            context._plugin.delete_l3_policy(context._plugin_context, l3p_id,
                                             check_unused=True)

    def _mark_l2_policy_owned(self, session, l2p_id):
        with session.begin(subtransactions=True):
            owned = OwnedL2Policy(l2_policy_id=l2p_id)
            session.add(owned)

    def _l2_policy_is_owned(self, session, l2p_id):
        with session.begin(subtransactions=True):
            return (session.query(OwnedL2Policy).
                    filter_by(l2_policy_id=l2p_id).
                    first() is not None)

    def _mark_l3_policy_owned(self, session, l3p_id):
        with session.begin(subtransactions=True):
            owned = OwnedL3Policy(l3_policy_id=l3p_id)
            session.add(owned)

    def _l3_policy_is_owned(self, session, l3p_id):
        with session.begin(subtransactions=True):
            return (session.query(OwnedL3Policy).
                    filter_by(l3_policy_id=l3p_id).
                    first() is not None)



# Copyright (c) 2014 Cisco Systems Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.common import constants as n_constants
from neutron.openstack.common import log
from neutron.plugins.ml2 import driver_api as api

from gbpservice.neutron.services.grouppolicy.drivers.cisco.apic import (
    apic_mapping as amap)

LOG = log.getLogger(__name__)


class APICMechanismGBPDriver(api.MechanismDriver):

    def initialize(self):
        self._apic_gbp = None

    @property
    def apic_gbp(self):
        if not self._apic_gbp:
            self._apic_gbp = (amap.ApicMappingDriver.
                              get_initialized_instance())
        return self._apic_gbp

    def create_port_postcommit(self, context):
        # DHCP Ports are created implicitly by Neutron, need to inform GBP
        if (context.current.get('device_owner') ==
                n_constants.DEVICE_OWNER_DHCP):
            self.apic_gbp.create_dhcp_policy_target_if_needed(
                context._plugin_context, context.current)

    def update_port_postcommit(self, context):
        self.apic_gbp.process_port_changed(context._plugin_context,
                                           context.original, context.current)

    def update_subnet_postcommit(self, context):
        self.apic_gbp.process_subnet_changed(context._plugin_context,
                                             context.original, context.current)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import requests

from neutron.openstack.common import log as logging
from oslo.config import cfg
from oslo.serialization import jsonutils
from requests import auth

LOG = logging.getLogger(__name__)
cfg.CONF.import_opt(
    'odl_username',
    'gbpservice.neutron.services.grouppolicy.drivers.odl.config',
    group='odl_driver'
)
cfg.CONF.import_opt(
    'odl_password',
    'gbpservice.neutron.services.grouppolicy.drivers.odl.config',
    group='odl_driver'
)
cfg.CONF.import_opt(
    'odl_host',
    'gbpservice.neutron.services.grouppolicy.drivers.odl.config',
    group='odl_driver'
)
cfg.CONF.import_opt(
    'odl_port',
    'gbpservice.neutron.services.grouppolicy.drivers.odl.config',
    group='odl_driver'
)


class OdlManager(object):
    """Class to manage ODL translations and workflow.

    This class manages translation from Neutron objects to APIC
    managed objects and contains workflows to implement these
    translations.
    """

    def __init__(self):

        LOG.info(_("Configured ODL username: %s"),
                 cfg.CONF.odl_driver.odl_username)
        LOG.info(_("Configured ODL password: %s"),
                 cfg.CONF.odl_driver.odl_password)
        LOG.info(_("Configured ODL host: %s"),
                 cfg.CONF.odl_driver.odl_host)
        LOG.info(_("Configured ODL port: %s"),
                 cfg.CONF.odl_driver.odl_port)

        self._username = cfg.CONF.odl_driver.odl_username
        self._password = cfg.CONF.odl_driver.odl_password
        self._host = cfg.CONF.odl_driver.odl_host
        self._port = cfg.CONF.odl_driver.odl_port
        self._headers = {
            'Content-type': 'application/yang.data+json',
            'Accept': 'application/yang.data+json',
        }

        self._base_url = (
            "http://%(host)s:%(port)s/restconf" %
            {'host': self._host, 'port': self._port}
        )
        self._reg_ep_url = (
            self._base_url +
            '/operations/endpoint:register-endpoint'
        )
        self._unreg_ep_url = (
            self._base_url +
            '/operations/endpoint:unregister-endpoint'
        )
        self._policy_url = (
            self._base_url +
            '/config/policy:tenants/policy:tenant/%(tenant_id)s'
        )
        self._action_url = (
            self._policy_url +
            '/subject-feature-instances/action-instance/%(action)s'
        )
        self._classifier_url = (
            self._policy_url +
            '/subject-feature-instances/classifier-instance/%(classifier)s'
        )
        self._l3ctx_url = (
            self._policy_url +
            '/l3-context/%(l3ctx)s'
        )
        self._l2bd_url = (
            self._policy_url +
            '/l2-bridge-domain/%(l2bd)s'
        )
        self._l2fd_url = (
            self._policy_url +
            '/l2-flood-domain/%(l2fd)s'
        )
        self._epg_url = (
            self._policy_url +
            '/policy:endpoint-group/%(epg)s'
        )
        self._subnet_url = (
            self._policy_url +
            '/subnet/%(subnet)s'
        )
        self._contract_url = (
            self._policy_url +
            '/policy:contract/%(contract)s'
        )

    def _convert2ascii(self, obj):
        if isinstance(obj, dict):
            return {self._convert2ascii(key): self._convert2ascii(value) for
                    key, value in obj.iteritems()}
        elif isinstance(obj, list):
            return [self._convert2ascii(element) for element in obj]
        elif isinstance(obj, unicode):
            return obj.encode('ascii', 'ignore')
        else:
            return obj

    def _sendjson(self, method, url, headers, obj=None):
        """Send json to the ODL controller."""

        medium = self._convert2ascii(obj) if obj else None
        url = self._convert2ascii(url)
        data = (
            jsonutils.dumps(medium, indent=4, sort_keys=True) if medium
            else None
        )
        LOG.debug("=========================================================")
        LOG.debug("Sending METHOD (%(method)s) URL (%(url)s)",
                  {'method': method, 'url': url})
        LOG.debug("(%(data)s)", {'data': data})
        LOG.debug("=========================================================")
        r = requests.request(
            method,
            url=url,
            headers=headers,
            data=data,
            auth=auth.HTTPBasicAuth(self._username,
                                    self._password)
        )
        r.raise_for_status()

    def _is_tenant_created(self, tenant_id):
        url = self._convert2ascii(self._policy_url % {'tenant_id': tenant_id})
        r = requests.request(
            'get',
            url=url,
            headers=self._headers,
            auth=auth.HTTPBasicAuth(self._username,
                                    self._password)
        )
        if r.status_code == 200:
            return True
        elif r.status_code == 404:
            return False
        else:
            r.raise_for_status()

    def register_endpoints(self, endpoints):
        for ep in endpoints:
            data = {"input": ep}
            self._sendjson('post', self._reg_ep_url, self._headers, data)

    def unregister_endpoints(self, endpoints):
        for ep in endpoints:
            data = {"input": ep}
            self._sendjson('post', self._unreg_ep_url, self._headers, data)

    def create_update_tenant(self, tenant_id, tenant):
        url = (self._policy_url % {'tenant_id': tenant_id})
        data = {"tenant": tenant}
        self._sendjson('put', url, self._headers, data)

    def create_action(self, tenant_id, action):
        """Create policy action"""
        self._touch_tenant(tenant_id)
        url = (self._action_url %
               {'tenant_id': tenant_id, 'action': action['name']})
        data = {"action-instance": action}
        self._sendjson('put', url, self._headers, data)

    def delete_action(self, tenant_id, action):
        """Delete policy action"""
        url = (self._action_url %
               {'tenant_id': tenant_id, 'action': action['name']})
        self._sendjson('delete', url, self._headers)

    def create_classifier(self, tenant_id, classifier):
        """Create policy classifier"""
        self._touch_tenant(tenant_id)
        url = (self._classifier_url %
               {'tenant_id': tenant_id, 'classifier': classifier['name']})
        data = {"classifier-instance": classifier}
        self._sendjson('put', url, self._headers, data)

    def delete_classifier(self, tenant_id, classifier):
        """Delete policy classifier"""
        url = (self._classifier_url %
               {'tenant_id': tenant_id, 'classifier': classifier['name']})
        self._sendjson('delete', url, self._headers)

    def create_update_l3_context(self, tenant_id, l3ctx):
        self._touch_tenant(tenant_id)
        url = (self._l3ctx_url %
               {'tenant_id': tenant_id, 'l3ctx': l3ctx['id']})
        data = {"l3-context": l3ctx}
        self._sendjson('put', url, self._headers, data)

    def delete_l3_context(self, tenant_id, l3ctx):
        url = (self._l3ctx_url %
               {'tenant_id': tenant_id, 'l3ctx': l3ctx['id']})
        self._sendjson('delete', url, self._headers)

    def create_update_l2_bridge_domain(self, tenant_id, l2bd):
        self._touch_tenant(tenant_id)
        url = (self._l2bd_url %
               {'tenant_id': tenant_id, 'l2bd': l2bd['id']})
        data = {"l2-bridge-domain": l2bd}
        self._sendjson('put', url, self._headers, data)

    def delete_l2_bridge_domain(self, tenant_id, l2bd):
        url = (self._l2bd_url %
               {'tenant_id': tenant_id, 'l2bd': l2bd['id']})
        self._sendjson('delete', url, self._headers)

    def create_update_l2_flood_domain(self, tenant_id, l2fd):
        self._touch_tenant(tenant_id)
        url = (self._l2fd_url %
               {'tenant_id': tenant_id, 'l2fd': l2fd['id']})
        data = {"l2-flood-domain": l2fd}
        self._sendjson('put', url, self._headers, data)

    def delete_l2_flood_domain(self, tenant_id, l2fd):
        url = (self._l2fd_url %
               {'tenant_id': tenant_id, 'l2fd': l2fd['id']})
        self._sendjson('delete', url, self._headers)

    def create_update_endpoint_group(self, tenant_id, epg):
        self._touch_tenant(tenant_id)
        url = (self._epg_url %
               {'tenant_id': tenant_id, 'epg': epg['id']})
        data = {"endpoint-group": epg}
        self._sendjson('put', url, self._headers, data)

    def delete_endpoint_group(self, tenant_id, epg):
        url = (self._epg_url %
               {'tenant_id': tenant_id, 'epg': epg['id']})
        self._sendjson('delete', url, self._headers)

    def create_update_subnet(self, tenant_id, subnet):
        self._touch_tenant(tenant_id)
        url = (self._subnet_url %
               {'tenant_id': tenant_id, 'subnet': subnet['id']})
        data = {"subnet": subnet}
        self._sendjson('put', url, self._headers, data)

    def delete_subnet(self, tenant_id, subnet):
        url = (self._subnet_url %
               {'tenant_id': tenant_id, 'subnet': subnet['id']})
        self._sendjson('delete', url, self._headers)

    def create_update_contract(self, tenant_id, contract):
        url = (self._contract_url %
               {'tenant_id': tenant_id, 'contract': contract['id']})
        data = {"contract": contract}
        self._sendjson('put', url, self._headers, data)

    def _touch_tenant(self, tenant_id):
        tenant = {
            "id": tenant_id
        }
        if not self._is_tenant_created(tenant_id):
            self.create_update_tenant(tenant_id, tenant)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from gbpservice.neutron.services.grouppolicy import (
    group_policy_driver_api as api)


class GroupPolicyContext(object):
    """GroupPolicy context base class."""
    def __init__(self, plugin, plugin_context):
        self._plugin = plugin
        self._plugin_context = plugin_context


class BaseResouceContext(GroupPolicyContext):
    def __init__(self, plugin, plugin_context, resource, original=None):
        super(BaseResouceContext, self).__init__(plugin, plugin_context)
        self._resource = resource
        self._original = original

    @property
    def current(self):
        return self._resource

    @property
    def original(self):
        return self._original


class PolicyTargetContext(GroupPolicyContext, api.PolicyTargetContext):

    def __init__(self, plugin, plugin_context, policy_target,
                 original_policy_target=None):
        super(PolicyTargetContext, self).__init__(plugin, plugin_context)
        self._policy_target = policy_target
        self._original_policy_target = original_policy_target

    @property
    def current(self):
        return self._policy_target

    @property
    def original(self):
        return self._original_policy_target

    def set_port_id(self, port_id):
        self._plugin._set_port_for_policy_target(
            self._plugin_context, self._policy_target['id'], port_id)
        self._policy_target['port_id'] = port_id


class PolicyTargetGroupContext(GroupPolicyContext,
                               api.PolicyTargetGroupContext):

    def __init__(self, plugin, plugin_context, policy_target_group,
                 original_policy_target_group=None):
        super(PolicyTargetGroupContext, self).__init__(plugin, plugin_context)
        self._policy_target_group = policy_target_group
        self._original_policy_target_group = original_policy_target_group

    @property
    def current(self):
        return self._policy_target_group

    @property
    def original(self):
        return self._original_policy_target_group

    def set_l2_policy_id(self, l2_policy_id):
        self._plugin._validate_shared_create(
                    self._plugin_context,
                    self._policy_target_group,
                    'policy_target_group')
        self._plugin._set_l2_policy_for_policy_target_group(
            self._plugin_context, self._policy_target_group['id'],
            l2_policy_id)
        self._policy_target_group['l2_policy_id'] = l2_policy_id

    def set_network_service_policy_id(self, network_service_policy_id):
        nsp_id = network_service_policy_id
        self._plugin._set_network_service_policy_for_policy_target_group(
            self._plugin_context, self._policy_target_group['id'], nsp_id)
        self._policy_target_group['network_service_policy_id'] = nsp_id

    def add_subnet(self, subnet_id):
        subnets = self._plugin._add_subnet_to_policy_target_group(
            self._plugin_context, self._policy_target_group['id'], subnet_id)
        self._policy_target_group['subnets'] = subnets


class L2PolicyContext(GroupPolicyContext, api.L2PolicyContext):

    def __init__(self, plugin, plugin_context, l2_policy,
                 original_l2_policy=None):
        super(L2PolicyContext, self).__init__(plugin, plugin_context)
        self._l2_policy = l2_policy
        self._original_l2_policy = original_l2_policy

    @property
    def current(self):
        return self._l2_policy

    @property
    def original(self):
        return self._original_l2_policy

    def set_l3_policy_id(self, l3_policy_id):
        self._plugin._validate_shared_create(
                    self._plugin_context,
                    self._l2_policy,
                    'l2_policy')
        self._plugin._set_l3_policy_for_l2_policy(
            self._plugin_context, self._l2_policy['id'], l3_policy_id)
        self._l2_policy['l3_policy_id'] = l3_policy_id

    def set_network_id(self, network_id):
        self._plugin._set_network_for_l2_policy(
            self._plugin_context, self._l2_policy['id'], network_id)
        self._l2_policy['network_id'] = network_id


class L3PolicyContext(GroupPolicyContext, api.L3PolicyContext):

    def __init__(self, plugin, plugin_context, l3_policy,
                 original_l3_policy=None):
        super(L3PolicyContext, self).__init__(plugin, plugin_context)
        self._l3_policy = l3_policy
        self._original_l3_policy = original_l3_policy

    @property
    def current(self):
        return self._l3_policy

    @property
    def original(self):
        return self._original_l3_policy

    def add_router(self, router_id):
        routers = self._plugin._add_router_to_l3_policy(
            self._plugin_context, self._l3_policy['id'], router_id)
        self._l3_policy['routers'] = routers

    def set_external_fixed_ips(self, external_segment_id, ips):
        self._l3_policy['external_segments'][external_segment_id] = ips
        self._plugin._update_ess_for_l3p(self._plugin_context,
                                         self._l3_policy['id'],
                                         self._l3_policy['external_segments'])

    def set_external_segment(self, external_segment_id):
        external_segments = {external_segment_id: []}
        self.current['external_segments'] = external_segments
        self._plugin.update_l3_policy(
            self._plugin_context, self.current['id'],
            {'l3_policy': {'external_segments': external_segments}})


class NetworkServicePolicyContext(
    GroupPolicyContext, api.NetworkServicePolicyContext):

    def __init__(self, plugin, plugin_context, network_service_policy,
                 original_network_service_policy=None):
        super(NetworkServicePolicyContext, self).__init__(
            plugin, plugin_context)
        self._network_service_policy = network_service_policy
        self._original_network_service_policy = original_network_service_policy

    @property
    def current(self):
        return self._network_service_policy

    @property
    def original(self):
        return self._original_network_service_policy


class PolicyClassifierContext(GroupPolicyContext, api.PolicyClassifierContext):

    def __init__(self, plugin, plugin_context, policy_classifier,
                 original_policy_classifier=None):
        super(PolicyClassifierContext, self).__init__(plugin, plugin_context)
        self._policy_classifier = policy_classifier
        self._original_policy_classifier = original_policy_classifier

    @property
    def current(self):
        return self._policy_classifier

    @property
    def original(self):
        return self._original_policy_classifier


class PolicyActionContext(GroupPolicyContext, api.PolicyActionContext):

    def __init__(self, plugin, plugin_context, policy_action,
                 original_policy_action=None):
        super(PolicyActionContext, self).__init__(plugin, plugin_context)
        self._policy_action = policy_action
        self._original_policy_action = original_policy_action

    @property
    def current(self):
        return self._policy_action

    @property
    def original(self):
        return self._original_policy_action


class PolicyRuleContext(GroupPolicyContext, api.PolicyRuleContext):

    def __init__(self, plugin, plugin_context, policy_rule,
                 original_policy_rule=None):
        super(PolicyRuleContext, self).__init__(plugin, plugin_context)
        self._policy_rule = policy_rule
        self._original_policy_rule = original_policy_rule

    @property
    def current(self):
        return self._policy_rule

    @property
    def original(self):
        return self._original_policy_rule


class PolicyRuleSetContext(GroupPolicyContext, api.PolicyRuleSetContext):

    def __init__(self, plugin, plugin_context, policy_rule_set,
                 original_policy_rule_set=None):
        super(PolicyRuleSetContext, self).__init__(plugin, plugin_context)
        self._policy_rule_set = policy_rule_set
        self._original_policy_rule_set = original_policy_rule_set

    @property
    def current(self):
        return self._policy_rule_set

    @property
    def original(self):
        return self._original_policy_rule_set


class ExternalSegmentContext(BaseResouceContext, api.ExternalSegmentContext):

    def add_subnet(self, subnet_id):
        self._plugin._set_subnet_to_es(self._plugin_context,
                                       self.current['id'], subnet_id)
        self.current['subnet_id'] = subnet_id


class ExternalPolicyContext(BaseResouceContext, api.ExternalPolicyContext):

    def set_external_segment(self, external_segment_id):
        external_segmets = [external_segment_id]
        self.current['external_segments'] = external_segmets
        self._plugin.update_external_policy(
            self._plugin_context, self.current['id'],
            {'external_policy': {'external_segments': external_segmets}})


class NatPoolContext(BaseResouceContext, api.NatPoolContext):
    pass
